/* This file contains the API for MD5.
 * Limitations of this implementation:
 *  -only handles octet boundaries (not bit oriented)
 *  -only handles 2^26 octets.
 *
 * To output an MD5 digest, the rfc says:
 *  The message digest produced as output is A, B, C, D. That is, we
 *  begin with the low-order byte of A, and end with the high-order byte
 *  of D.
 *
 */
typedef struct md5_digest_t {
    uint32_t a;
    uint32_t b;
    uint32_t c;
    uint32_t d;
} md5_digest;

/* Get the size of the md5 digest
 */
int md5_digest_len();

/* Returns an initial digest to begin computation */
md5_digest new_md5_digest();

/* Compute the md5 of an arbitrary length (in octets) message */
md5_digest md5(uint8_t *m, uint64_t len);

/* The MD5 compression function computes the md5 of a 512-bit quantity.
 * It assumes m is at least 512 bits long, and d is the result
 *  of the precediing 512 bits.
 * Shorter messages must be properly padded, and the first 512-bits must
 *  be computed in relation to new_md5().
 * d is modified to the new value and is also returned;
 */
md5_digest md5_compress(uint8_t *m, md5_digest *d);


/* An md5 stream allows computation of a digest for a message
 *  from arbitrarily sized continuous portions.
 */

typedef struct md5_stream_context_t {
    md5_digest d;
    uint64_t total_computed;     /* in octets */
    unsigned int buf_used;
    uint8_t buf[64];
} md5_stream_context;

/* Initialize a context for a stream */
md5_stream_context md5_stream_init();

/* Process an arbitrary number of octets */
void md5_stream(uint8_t *message, uint64_t len, md5_stream_context *context);

/* Compute result of stream */
md5_digest md5_stream_result(md5_stream_context *context);

/* Convert a digest into a string
 * Assumes str has already been alocated at least 33 characters.
 * Returns str.
 */
char *md5_digest2str(md5_digest *d, char *str);

/* Copy md5 digest to dest in network byte order
 */
void md5_digest_copy(uint8_t *dest, md5_digest *d);
