#include <conf.h>
#include <kernel.h>
#include <network.h>
#include <ipopt.h>
#include <bufpool.h>
#include <mq.h>
#include <timer.h>
#include <stdio.h>

/* #define DEBUG */

struct ipfc ipfctab[IPFC_N];

extern int ctr1000;

inline static int ipeq (struct ip *pip1, struct ip *pip2)
{
	return (pip1->ip_src == pip2->ip_src
		&& pip1->ip_dst == pip2->ip_dst
		&& pip1->ip_id == pip2->ip_id
		&& pip1->ip_proto == pip2->ip_proto);
}

static void ipfccoallesce (struct ep *pep1, struct ip *pip1,
	struct ep *pep2, struct ip *pip2)
{
	int nlen;
	u_short offset;
	nlen = net2hs (pip1->ip_len) + net2hs(pip2->ip_len) - IP_HLEN (pip2);
	if (IP_OFF(pip1) <= IP_OFF(pip2)
	    && IP_OFF(pip1) + net2hs(pip1->ip_len)
	      - IP_HLEN(pip1) >= IP_OFF(pip2)
	    && sizeof(struct ehx) + sizeof(struct eh) + nlen <= IP_SMLBUFSZ) {
		blkcopy (pep1->ep_data + IP_HLEN(pip1) + IP_OFF(pip2) - IP_OFF(pip1),
			pep2->ep_data + IP_HLEN(pip2),
			net2hs(pip2->ip_len) - IP_HLEN(pip2));
		pip1->ip_len = hs2net (nlen);
		offset = net2hs (pip1->ip_offset);
		offset &= ~(IPF_MF);
		offset |= net2hs (pip2->ip_offset) & IPF_MF;
		pip1->ip_offset = hs2net (offset);
		pip1->ip_cksum = 0;
		pip1->ip_cksum = cksum (pip1, IP_HLEN(pip1));
		pep1->ep_len = nlen + sizeof(struct eh);
		pep1->ep_next = pep2->ep_next;
		Ip.nfrags--;
		freebuf (pep2);
	}
}

/* assume sorted by last byte, at least one entry */
/* packet complete when !(hole & MF) */
static int ipfcadd (struct ipfc *pfc, struct ep *pep, struct ip *pip)
{
	struct ep *cur, *prev;
	struct ip *curip, *nextip, *holeip;
#ifdef DEBUG
	kprintf("ipfc adding 0x%x\n", (u_int)pep);
#endif
	if (pfc->hole == NULL) {
		cur = pfc->head;
		prev = NULL;
	} else {
		holeip = (struct ip *) pfc->hole->ep_data;
		if (IP_OFF(holeip) + net2hs(holeip->ip_len)
		    >= IP_OFF(pip) + net2hs(pip->ip_len)) {
			// if it doesn't help us, free it.
			freebuf (pep);
			return OK;
		}
		cur = pfc->hole->ep_next;
		prev = pfc->hole;
	}
	while (cur != NULL) {
		curip = (struct ip*) cur->ep_data;
		if (IP_OFF(pip) + net2hs(pip->ip_len)
		    < IP_OFF(curip) + net2hs(curip->ip_len)) {
			pep->ep_next = cur;
			if (prev == NULL)
				pfc->head = pep;
			else
				prev->ep_next = pep;
			ipfccoallesce (pep, pip, cur, curip);
			if (prev != NULL)
				ipfccoallesce (prev, (struct ip*)prev->ep_data,
					pep, pip);
			break;
		}
		prev = cur;
		cur = cur->ep_next;
	}
	if (cur == NULL) {
		if (prev == NULL)
			pfc->head = pep;
		else {
			prev->ep_next = pep;
			ipfccoallesce (prev, (struct ip*)prev->ep_data, pep, pip);
		}
	}
	
	// recalculate hole.
	if (IP_OFF((struct ip*)pfc->head->ep_data) != 0) {
		pfc->hole = NULL;
	} else {
		cur = pfc->hole;
		if (cur == NULL)
			cur = pfc->head;
		curip = (struct ip *)cur->ep_data;
		while (cur->ep_next != NULL) {
			nextip = (struct ip *)cur->ep_next->ep_data;
			if (IP_OFF(curip) + net2hs(curip->ip_len)
			    <= IP_OFF(nextip))
				break;
			cur = cur->ep_next;
			curip = nextip;
		}
		pfc->hole = cur;
	}
	return OK;
}

static void ipfcclear (struct ipfc *pfc)
{
	struct ep *cur, *next;
	cur = pfc->head;
	while (cur != NULL) {
		next = cur->ep_next;
		freebuf (cur);
		cur = next;
		Ip.nfrags--;
	}
	pfc->head = NULL;
}

/* Remove the packet which has been least recently updated */
static void ipfcclearlru ()
{
	int i, lru, lruts;

	lruts = ctr1000 + 1;
	lru = 0;
	for (i = 0; i < IPFC_N; i++) {
		if (ipfctab[i].head != NULL && ipfctab[i].rcntts - lruts < 0)
			lruts = ipfctab[lru = i].rcntts;
	}
	ipfcclear (&ipfctab[lru]);
}

inline static void ipfcalloc (struct ipfc *pfc, struct ep *pep)
{
	ipfcclear (pfc);
	pfc->head = pep;
	pfc->hole = NULL;
	pfc->firstts = ctr1000;
	pfc->rcntts = ctr1000;
	tmset (IPFC_TIMEOUT + 5, Ip.slowipmq, NULL);
}

/* Construct a packet from fragments */
/* (delivers packet) */
inline static void ipfcassemble (struct ipfc *pfc)
{
	struct ep *cur, *npep;
	struct ip *curip, *npip, *headip, *holeip;
	int len;

	headip = (struct ip*)pfc->head->ep_data;
	holeip = (struct ip*)pfc->hole->ep_data;
#ifdef DEBUG
	kprintf("Packet complete\n");
#endif
	len = IP_OFF(holeip) + net2hs(holeip->ip_len) - IP_HLEN(holeip);
	npep = _ipgetbuf(len + IP_HLEN(headip) +
		sizeof(struct eh) + sizeof(struct ehx), FALSE);
	if (npep == NULL || npep == (struct ep*)SYSERR) {
		ipfcclear (pfc);
		return;
	}
		
	blkcopy (npep, pfc->head, sizeof(struct eh) + sizeof(struct ehx)
		+ net2hs(headip->ip_len));
	npip = (struct ip*) npep->ep_data;
	cur = pfc->head->ep_next;
	while (cur != NULL) {
		curip = (struct ip*)cur->ep_data;
		blkcopy (npep->ep_data + IP_HLEN(npip) + IP_OFF(curip),
			(u_char*)curip + IP_HLEN(curip),
			net2hs(curip->ip_len) - IP_HLEN(curip));
		cur = cur->ep_next;
	}
	ipfcclear (pfc);
	npip->ip_len = hs2net(IP_HLEN(npip) + len);
	npip->ip_offset = 0;
	npip->ip_cksum = 0;
	npip->ip_cksum = cksum (npip, IP_HLEN(npip));
	npep->ep_next = NULL;
	npep->ep_len = IP_HLEN(npip) + len + sizeof(struct eh);
	ipdeliver (npep, npip);
}

void ipreassemble (struct ep *pep, struct ip *pip)
{
	int i, lru, lruts;

#ifdef DEBUG
	kprintf("ipreassemble 0x%x off=%d dlen=%d\n", (u_int)pep,
		(int) IP_OFF(pip), (int)(net2hs(pip->ip_len) - IP_HLEN(pip)));
#endif
	if (Ip.nfrags >= IPFC_FRAGMAX)
		ipfcclearlru();
	Ip.nfrags++;

	pep->ep_next = NULL;
	for (i = 0; i < IPFC_N; i++) {
		if (ipfctab[i].head != NULL
		    && (ipeq(pip, (struct ip*) ipfctab[i].head->ep_data))) {
			ipfcadd (&ipfctab[i], pep, pip);
			if (ipfctab[i].hole != NULL && !(net2hs(((struct ip*)
			    (ipfctab[i].hole->ep_data))->ip_offset)
			    & IPF_MF)) {
				ipfcassemble (&ipfctab[i]);
			} else {
			}
			ipfctab[i].rcntts = ctr1000;
			break;
		}
	}
	if (i == IPFC_N) { /* Create a new entry */
		lruts = ctr1000 + 1;
		lru = 0;
		for (i = 0; i < IPFC_N; i++) {
			if (ipfctab[i].rcntts - lruts < 0)
				lruts = ipfctab[lru = i].rcntts;
			if (ipfctab[i].head == NULL) {
				ipfcalloc(&ipfctab[i], pep);
				break;
			}
		}
		if (i == IPFC_N)
			ipfcalloc(&ipfctab[lru], pep);
	}
	
	return;
}

/* For debugging */
int ipprintfc (int stdout)
{
	struct ep *cur;
	struct ip *curip;
	int i;
	struct ipfc *pfc;
	for (i = 0; i < IPFC_N; i++) {
		pfc = &ipfctab[i];
		fprintf (stdout,"\n[%d] head=0x%x, hole=0x%x, fts=%d, rts=%d\n",
			i, (u_int)pfc->head, (u_int)pfc->hole,
			pfc->firstts, pfc->rcntts);
		cur = pfc->head;
		while (cur != NULL) {
			curip = (struct ip*) cur->ep_data;
			fprintf (stdout, " [0x%x : %d-%d]", (u_int)cur,
				(int)IP_OFF(curip), (int)(IP_OFF(curip)
				+ net2hs(curip->ip_len) - IP_HLEN(curip)));
			cur = cur->ep_next;
		}
	}
	fprintf (stdout, "\n");
	
	return OK;
}

/* Remove all entries which have timed out */
void ipfctimeout ()
{
	int i;
	struct ipfc *pfc;

	for (i = 0; i < IPFC_N; i++) {
		if ((pfc = &ipfctab[i])->head != NULL
		    && pfc->firstts - (ctr1000 - IPFC_TIMEOUT) < 0) {
			if (IP_OFF((struct ip*) pfc->head->ep_data) == 0)
				icmpreply (ICMPT_TIMEX, ICMPC_FRAG, 0, pfc->head);
			ipfcclear (pfc);
		}
	}
}
