#include <conf.h>
#include <kernel.h>
#include <iproute.h>

struct route *rtnew (IPaddr prefix, IPaddr mask, IPaddr gateway,
        u_int ifnum, u_int metric, u_int ttl, u_int tag, u_int notadv);

int rtadd (IPaddr prefix, IPaddr mask, IPaddr gateway, u_int ifnum,
                u_int metric, u_int ttl, u_int tag, u_int notadv) {
        int rv;

        wait (Route.ri_wsem);
        rv = rtaddnl (prefix, mask, gateway, ifnum,
               metric, ttl, tag, notadv);
        signal(Route.ri_wsem);
        rtcommit();
        return rv;
}

int rtaddnl (IPaddr prefix, IPaddr mask, IPaddr gateway, u_int ifnum,
                u_int metric, u_int ttl, u_int tag, u_int notadv)
{
        int isdup;
        struct rttable *prtbl;
        struct route *prev, *srt, *prt;
        struct route *tmprt;

        prtbl = Route.ri_working;

        prt = rtnew(prefix, mask, gateway, ifnum, metric, ttl, tag, notadv);
        if (prt == (struct route *)SYSERR) {
                return(SYSERR);
        }

        prev = NULL;
        isdup = FALSE;
        for(srt=RT_PHEAD(prtbl); srt; srt=RT_PNEXT(prtbl, srt)) {
                if(RT_ISPREFIXOF(srt, prt)) prt->rt_iprefix=srt->rt_iself;

                if (net2hl(prt->rt_net) < net2hl(srt->rt_net) ||
                        (prt->rt_net == srt->rt_net &&
                         net2hl(prt->rt_mask) < net2hl(srt->rt_mask)))
                                break;
                
                if (prt->rt_net == srt->rt_net &&
                    prt->rt_mask == prt->rt_mask) {
                        isdup = TRUE;
                        break;
                }
                
                prev = srt;
        }

        if(isdup) {
                if (srt->rt_gw == prt->rt_gw) {
                        srt->rt_metric = metric;
                        srt->rt_ttl = ttl;
                        srt->rt_tag = tag;
                        srt->rt_flags |= RTF_CHANGED;
                        RT_PFREE(prtbl, prt);
                        return OK;
                }

                if(srt->rt_metric <= prt->rt_metric) {
                        /* no better off to change; drop the new one */
                        RT_PFREE(prtbl, prt);
                        return OK;
                }

                tmprt = srt;
                srt = RT_PNEXT(prtbl, srt);
                prt->rt_flags |= TRUE;
                RT_PFREE(prtbl, tmprt);
        }

        /* this needs to be optimized */
        prt->rt_inext = RT_IPTR(srt);
        for(;srt;srt=RT_PNEXT(prtbl, srt)) {
                if(RT_ISPREFIXOF(prt, srt)) {
                        if(srt->rt_iprefix==prt->rt_iprefix || srt->rt_iprefix==RT_IEND)
                                srt->rt_iprefix = prt->rt_iself;
                }
        }

        if (prev)
                prev->rt_inext = prt->rt_iself;
        else
                prtbl->rtb_irlist = prt->rt_iself;

        return OK;

}

void rtcommit() {
        int    rcount;
        struct rttable *tmp_prtable;

        wait (Route.ri_wsem);
        rttrie(); 
        wait (Route.ri_rsem);
        tmp_prtable = Route.ri_lookup;
        Route.ri_lookup = Route.ri_working;
        Route.ri_working = tmp_prtable;        
        rcount = Route.ri_working->rtb_rcount;
        signal (Route.ri_rsem);
        if (rcount) wait (Route.ri_refsem);
        blkcopy (Route.ri_working, Route.ri_lookup, sizeof(struct rttable));
        signal (Route.ri_wsem);
}

struct route *rtnew(IPaddr prefix, IPaddr mask, IPaddr gateway,
                 u_int ifnum, u_int metric, u_int ttl, u_int tag, u_int notadv)
{
        struct route *prt;
        
        if(Route.ri_working->rtb_ifree == RT_IEND)
                return (struct route *)SYSERR;

        prt = RT_PTR(Route.ri_working, Route.ri_working->rtb_ifree);
        Route.ri_working->rtb_ifree = prt->rt_inext;

        prt->rt_inext = RT_IEND;
        prt->rt_iprefix = RT_IEND;

        prt->rt_net = prefix & mask;
        prt->rt_mask = mask;
        prt->rt_gw = gateway;
        prt->rt_ifnum = ifnum;
        prt->rt_metric = metric;
        prt->rt_ttl = ttl;
        prt->rt_tag = tag;
        prt->rt_flags = RTF_NULL;
        if (notadv) prt->rt_flags |= RTF_NOTADV;

        return prt;
}

